SECTION "SIAL-386w"

/*	This program compiles .sial files to .s files for 386 machines.
	Implemented by Martin Richards (c) August 2014

Change history

26/02/2021
	This version of "SIAL-386w" has been revised to bring the code generated
	into line with code generated by Robert Nordier's 'xg' program, which
	converts INTCODE to Intel-386 assembler.

	SIAL-386w's resulting assembler files may be assembled and linked with
	object and library modules forming part of the BCPLkit distribution
	
															Dave Cannon

21/08/2014
	Added some simple optimisations to the floating point operations.

20/08/2014
	Adding floating point operations
	FIX, FLOAT, #ABS, #*, #/, #+, #-, #=, #~=, #<=, #>=, #< and #>. 
*/


GET "libhdr"

GET "sial.h"

/*
The following extracts from: "The BCPL Cintsys and Cintpos User Guide"
			by Martin Richards
		
The prefixes are as follows:

	F An opcode or directive
	P A stack offset, 0 to #xFFFFFF
	G A global variable number, 0 to 65535
	K A 24-bit unsigned constant, often small in value
	W A 32-bit signed integer, used for static data and large constants
	C A byte in range 0 to 255
	L A label generated by translation phase
	M A label generated by the Sial codegenerator
	
The instructions are for an abstract machine with internal registers:

	a The main accumulator, function first arg and result register
	b The second accumulator used in dyadic operations
	c Register used by pbyt and xpbyt, and possibly corrupted by
	  some other instructions, such as mul, div, rem, xdiv and xrem
	P Pointer to the base of the current stack frame
	G Pointer to the base of the Global Vector
	PC Set by jump and call instructions

When implementing sial-386 it was necessary to decide how the Intel registers
were to be used and what the BCPL calling sequence should be. The chosen register
allocation was as follows:

	Intel register | Use							| DC: Revised usage for SIAL-386w
	_______________|________________________________|______________
		%ebx	   | The A register					|	%eax
		%ecx	   | The B register					|	%ebx
		%edx	   | The C register					|	%ecx
		%eax	   | A work register				|	%edx
		%esi	   | Pointer to the GLOBAL vector	|	%edi
		%edi	   | A work register				|	%esi
		%ebp	   | The stackframe (P) pointer		|	%ebp
*/

MANIFEST {
  k_none=0
  k_a
  k_b
  k_f     // Floating point accumulator, typically st(0)
  k_loc   // Local variable Pn
  k_glob  // Global variable Gn
  k_lab   // Static locations addressed by Ln
  k_mlab  // Static locations addressed by Mn
}

GLOBAL {
sialin: ug
asmout; stdin; stdout

rdf; rdp; rdg; rdk; rdw; rdl; rdc
rdcode

pval; gval; kval; wval; lval; mval

scan
cvf; cvfp; cvfg; cvfk; cvfw; cvfl

sectname
modletter
charv
labnumber

regb_k; regb_n // Describes what should be in the left operand
               // k_a
               // k_b
               // k_f      the floating point accumulator
               // k_loc    local n
               // k_glob   global n
               // k_lab    static location labelled Ln
               // k_mlab   static location labelled Mn

rega_k; rega_n // Describes what should be in the right operand

tracing
prstate

loada  // (k, n)
loadb  // (k, n)
}

LET start() = VALOF
{ LET argv = VEC 20
  LET v    = VEC 20
  LET cv   = VEC 256/bytesperword

  tracing := TRUE
  sectname := v
  sectname%0 := 0
  modletter := 'A'
  charv := cv
  labnumber := 0

  asmout := 0
  stdout := output()
  IF rdargs("FROM,TO/K", argv, 20)=0 DO
  { writes("Bad args for sial-386*n")
    RESULTIS 20
  }
  IF argv!0=0 DO argv!0 := "prog.sial"
  IF argv!1=0 DO argv!1 := "prog.s"
  sialin := findinput(argv!0)
  IF sialin=0 DO
  { writef("Trouble with file %s*n", argv!0)
    RESULTIS 20
  }
  asmout := findoutput(argv!1)
   
  IF asmout=0 DO
  { writef("Trouble with file %s*n", argv!1)
    RESULTIS 20
  }
   
  writef("SIAL-386w (8-Sep-2014)*nConverting %s to %s*n", argv!0, argv!1)
  selectinput(sialin)
  selectoutput(asmout)

  // Initialise the state
  regb_k, regb_n := k_b, 0
  rega_k, rega_n := k_a, 0

  writef("# Code generated by SIAL-386w (8-Sep-2014)*n")
  writef("*n.align 4,0*n.text")

  scan()
  endread()
  UNLESS asmout=stdout DO endwrite()
  selectoutput(stdout)
  writef("Conversion complete*n")
  RESULTIS 0
}

AND nextlab() = VALOF
{ labnumber := labnumber+1
  RESULTIS labnumber
}

// argument may be of form Ln
AND rdcode(let) = VALOF
{ LET a, ch, neg = 0, ?, FALSE

  { ch := rdch()
    IF ch='/' DO ch := rdch() REPEATUNTIL ch='*n' | ch=endstreamch	// permit comments beginning '/'...
  } REPEATWHILE ch='*s' | ch='*n'

  IF ch=endstreamch RESULTIS -1
  UNLESS ch=let DO error("Bad item, looking for %c found %c*n", let, ch)
  ch := rdch()
  IF ch='-' DO { neg := TRUE; ch := rdch() }
  WHILE '0'<=ch<='9' DO { a := 10*a - (ch - '0'); ch := rdch()  }
  RESULTIS neg -> a, -a
}

AND rdf() = rdcode('F')
AND rdp() = VALOF { pval := rdcode('P'); RESULTIS pval }
AND rdg() = VALOF { gval := rdcode('G'); RESULTIS gval }
AND rdk() = VALOF { kval := rdcode('K'); RESULTIS kval }
AND rdw() = VALOF { wval := rdcode('W'); RESULTIS wval }
AND rdl() = VALOF { lval := rdcode('L'); RESULTIS lval }
AND rdm() = VALOF { mval := rdcode('M'); RESULTIS mval }
AND rdc() = rdcode('C')

AND error(mess, a, b, c) BE
{ LET out = output()
  UNLESS out=stdout DO
  { selectoutput(stdout)
    writef(mess, a, b, c)
    selectoutput(out)
  }
  writef(mess, a, b, c)
}

AND prstate() BE IF tracing DO
{ writef("	# RegB = ")
  SWITCHON regb_k INTO
  { DEFAULT:     writef("?");            ENDCASE
    CASE k_none: writef("-");            ENDCASE
    CASE k_a:    writef("A");            ENDCASE
    CASE k_b:    writef("B");            ENDCASE
    CASE k_f:    writef("F");            ENDCASE
    CASE k_loc:  writef("P%n",  regb_n); ENDCASE
    CASE k_glob: writef("G%n",  regb_n); ENDCASE
    CASE k_lab:  writef("L%n",  regb_n); ENDCASE
    CASE k_mlab: writef("M%n",  regb_n); ENDCASE
  }
  writef("	#  RegA = ")
  SWITCHON rega_k INTO
  { DEFAULT:     writef("?");            ENDCASE
    CASE k_none: writef("-");            ENDCASE
    CASE k_a:    writef("A");            ENDCASE
    CASE k_b:    writef("B");            ENDCASE
    CASE k_f:    writef("F");            ENDCASE
    CASE k_loc:  writef("P%n",  rega_n); ENDCASE
    CASE k_glob: writef("G%n",  rega_n); ENDCASE
    CASE k_lab:  writef("L%n",  rega_n); ENDCASE
    CASE k_mlab: writef("M%n",  rega_n); ENDCASE
  }
  newline()
}

AND loada() BE
{ // Compile code to ensure that reg A holds it true value
  // ie that rega_k=k_a

  SWITCHON rega_k INTO
  { DEFAULT:
      writef("loada error: bad k %n*n", rega_k)
      abort(999)
      ENDCASE
    CASE k_a:  // Already in A
      ENDCASE
    CASE k_b:
      TEST regb_k=k_a
      THEN { // Must exchange A and B
             writef("*n	xchgl %%ebx,%%eax")
             regb_k := k_b
           }
      ELSE { // Just move B into A
             writef("*n	movl %%ebx,%%eax")
           }
      ENDCASE
    CASE k_f:
      writeq("*n# F -> A")
      ENDCASE
    CASE k_loc:
      writef("*n	movl %n(%%ebp),%%eax", 4*rega_n)
      ENDCASE
    CASE k_glob:
      writef("*n	movl %n(%%edi),%%eax", 4*rega_n)
      ENDCASE
    CASE k_lab:
      writef("*n	movl L%c%n,%%eax", modletter, rega_n)
      ENDCASE
    CASE k_mlab:
      writef("*n	movl M%c%n,%%eax", modletter, rega_n)
      ENDCASE
  }
  rega_k, rega_n := k_a, 0
}

AND freea() BE
{ IF regb_k=k_a DO
  { // Must move A into B
    writef("*n	movl %%eax,%%ebx")
    regb_k := k_b
  }
  rega_k := k_a
}

AND loadb() BE
{ // Compile code to ensure that reg B holds it true value
  // ie that regb_k=k_b

  prstate()
  SWITCHON regb_k INTO
  { DEFAULT:
      writef("*n*nloadb error: bad k %n*n", regb_k)
      abort(999)
      ENDCASE
    CASE k_a: // The true value of reg b is in A
      TEST rega_k=k_b
      THEN { // Must exchange A and B
             writef("*n	xchgl %%eax,%%ebx")
             rega_k := k_a
           }
      ELSE { // Just move A into B
             writef("*n	movl %%eax,%%ebx")
           }
      ENDCASE
    CASE k_b:
      ENDCASE // Already in B
    CASE k_f:
      writeq("*n# F -> B")
      ENDCASE
    CASE k_loc:
      writef("*n	movl %n(%%ebp),%%ebx", 4*regb_n)
      ENDCASE
    CASE k_glob:
      writef("*n	movl %n(%%edi),%%ebx", 4*regb_n)
      ENDCASE
    CASE k_lab:
      writef("*n	movl L%c%n,%%ebx", modletter, regb_n)
      ENDCASE
    CASE k_mlab:
      writef("*n	movl M%c%n,%%ebx", modletter, regb_n)
      ENDCASE
  }
  regb_k, regb_n := k_b, 0
}

AND freeb() BE
{ IF rega_k=k_b DO
  { // Must move B into A
    writef("*n	movl %%ebx,%%eax")
    rega_k := k_a
  }
  regb_k := k_b
}

AND scan() BE
{ LET op = rdf()

  SWITCHON op INTO

  { DEFAULT:       error("# Bad op %n*n", op); LOOP

    CASE -1:       RETURN

    CASE f_lp:     cvfp("LP") // a := P!n
                   writef("*n	movl %n(%%ebp),%%eax", 4*pval)
                   ENDCASE
    CASE f_lg:     cvfg("LG") // a := G!n
                   writef("*n	movl %n(%%edi),%%eax", 4*gval)
                   ENDCASE
    CASE f_ll:     cvfl("LL") // a := !Ln
                   writef("*n	movl L%c%n,%%eax", modletter, lval)
                   ENDCASE
    CASE f_llp:    cvfp("LLP") // a := @ P!n
                   writef("*n	leal %n(%%ebp),%%eax", 4*pval)
                   writef("*n	shrl $2,%%eax")
                   ENDCASE
    CASE f_llg:    cvfg("LLG") // a := @ G!n
                   writef("*n	leal %n(%%edi),%%eax", 4*gval)
                   writef("*n	shrl $2,%%eax")
                   ENDCASE
    CASE f_lll:    cvfl("LLL") // a := @ !Ln
                   writef("*n	movl $L%c%n,%%eax", modletter, lval)
                   writef("*n	shrl $2,%%eax")
                   ENDCASE
    CASE f_lf:     cvfl("LF") // a := byte address of Ln
                   writef("*n	leal L%c%n,%%eax", modletter, lval)
                   ENDCASE
    CASE f_lw:     cvfm("LW")
                   writef("*n	movl M%c%n,%%eax", modletter, mval)
                   ENDCASE
    CASE f_l:      cvfk("L") // a := n
                   IF kval=0 DO { writef("*n	xorl %%eax,%%eax"); ENDCASE }
                   writef("*n	movl $%n,%%eax", kval)
                   ENDCASE
    CASE f_lm:     cvfk("LM") // a := -n
                   writef("*n	movl $-%n,%%eax", kval)
                   ENDCASE
    CASE f_sp:     cvfp("SP") // P!n := a
                   writef("*n	movl %%eax,%n(%%ebp)", 4*pval)
                   ENDCASE
    CASE f_sg:     cvfg("SG") // G!n := a
                   writef("*n	movl %%eax,%n(%%edi)", 4*gval)
                   ENDCASE
    CASE f_sl:     cvfl("SL") // !Ln := a
                   writef("*n	movl %%eax,L%c%n", modletter, lval)
                   ENDCASE
    CASE f_ap:     cvfp("AP") // a := a + P!n
                   writef("*n	addl %n(%%ebp),%%eax", 4*pval)
                   ENDCASE
    CASE f_ag:     cvfg("AG") // a := a + G!n
                   writef("*n	addl %n(%%edi),%%eax", 4*gval)
                   ENDCASE
    CASE f_a:      cvfk("A") // a := a + n
                   IF kval=0 ENDCASE
                   IF kval=1  DO { writef("*n	incl %%eax"); ENDCASE }
                   IF kval=-1 DO { writef("*n	decl %%eax"); ENDCASE }
                   writef("*n	addl $%n,%%eax", kval)
                   ENDCASE
    CASE f_s:      cvfk("S")  // a := a - n
                   IF kval=0 ENDCASE
                   IF kval=1  DO { writef("*n	decl %%eax"); ENDCASE }
                   IF kval=-1 DO { writef("*n	incl %%eax"); ENDCASE }
                   writef("*n	subl $%n,%%eax", kval)
                   ENDCASE
    CASE f_lkp:    cvfkp("LKP") // a := P!n!k
                   writef("*n	movl %n(%%ebp),%%edx", 4*pval)
                   writef("*n	movl %n(,%%edx,4),%%eax", 4*kval)
                   ENDCASE
    CASE f_lkg:    cvfkg("LKG") // a := G!n!k
                   writef("*n	movl %n(%%edi),%%edx", 4*gval)
                   writef("*n	movl %n(,%%edx,4),%%eax", 4*kval)
                   ENDCASE
    CASE f_rv:     cvf("RV")  // a := ! a
                   writef("*n	movl (,%%eax,4),%%eax")
                   ENDCASE
    CASE f_rvp:    cvfp("RVP") // a := P!n!a
                   writef("*n	addl %n(%%ebp),%%eax", 4*pval)
                   writef("*n	movl (,%%eax,4),%%eax")
                   ENDCASE
    CASE f_rvk:    cvfk("RVK") // a := a!k
                   writef("*n	movl %n(,%%eax,4),%%eax", 4*kval)
                   ENDCASE
    CASE f_st:     cvf("ST") // !a := b
                   writef("*n	movl %%ebx,(,%%eax,4)")
                   ENDCASE
    CASE f_stp:    cvfp("STP") // P!n!a := b
                   writef("*n	movl %n(%%ebp),%%edx", 4*pval)
                   writef("*n	addl %%eax,%%edx")
                   writef("*n	movl %%ebx,(,%%edx,4)")
                   ENDCASE
    CASE f_stk:    cvfk("STK") // a!n := b
                   writef("*n	movl %%ebx,%n(,%%eax,4)", 4*kval)
                   ENDCASE
    CASE f_stkp:   cvfkp("STKP")  // P!n!k := a
                   writef("*n	movl %n(%%ebp),%%edx", 4*pval)
                   writef("*n	movl %%eax,%n(,%%edx,4)", 4*kval)
                   ENDCASE
    CASE f_skg:    cvfkg("SKG") // G!n!k := a
                   writef("*n	movl %n(%%edi),%%edx", 4*gval)
                   writef("*n	movl %%eax,%n(,%%edx,4)", 4*kval)
                   ENDCASE
    CASE f_xst:    cvf("XST") // !b := a
                   writef("*n	movl %%eax,(,%%ebx,4)")
                   ENDCASE
    CASE f_k:      cvfp("K") // Call  a(b,...) incrementing P by n
              /*   writef("*n	leal %n(%%ebp),%%ecx", 4*pval)
               *   writef("*n	calll **%%eax")
               *   writef("*n	movl %n(%%edi),%%eax", 4*pval)
               */
				   writef("*n	movl %%ebp,%%ecx")
				   writef("*n	addl $%n,%%ebp", 4*pval)
				   writef("*n	movl %%ecx,(%%ebp)")
				   writef("*n	movl $1f,4(%%ebp)")
				   writef("*n	jmp **%%eax")
				   writef("*n1:")

                   ENDCASE
    CASE f_kpg:    cvfpg("KPG") // Call Gg(a,...) incrementing P by n
              /*   writef("*n	movl %%eax,%n(%%ebp)", 4*(pval+3)) 			// pre-store arg2 - needed for non-sial compiled libraries and OS interface
               *   writef("*n	leal %n(%%ebp),%%ecx", 4*pval)
               *   writef("*n	calll **%n(%%edi)", 4*gval)
               */
                   writef("*n	movl %%eax,%n(%%ebp)", 4*(pval+3)) 			// pre-store arg2 - needed for non-sial compiled libraries and OS interface
                   writef("*n	movl %n(%%edi),%%eax", 4*gval)
				   writef("*n	movl %%ebp,%%ecx")
				   writef("*n	addl $%n,%%ebp", 4*pval)
				   writef("*n	movl %%ecx,(%%ebp)")
				   writef("*n	movl $1f,4(%%ebp)")
				   writef("*n	jmp **%%eax")
				   writef("*n1:")

                   ENDCASE
    CASE f_neg:    cvf("NEG") // a := - a
                   writef("*n	negl %%eax") 
                   ENDCASE
    CASE f_not:    cvf("NOT") // a := ~ a
                   writef("*n	notl %%eax") 
                   ENDCASE
    CASE f_abs:    cvf("ABS") // a := ABS a
                 { LET l = nextlab()
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	jge L%n", l)
                   writef("*n	negl %%eax")
                   writef("*nL%n:", l)
                   ENDCASE
                 }
    CASE f_xdiv:   cvf("XDIV") // a := a / b
                   writef("*n	cltd")
                   writef("*n	idivl %%ebx")
                   ENDCASE
    CASE f_xrem:   cvf("XREM") // a := a REM b
                   writef("*n	cltd")
                   writef("*n	idivl %%ebx")
                   writef("*n	movl %%edx,%%eax")
                   ENDCASE
    CASE f_xsub:   cvf("XSUB") // a := a - b
                   writef("*n	subl %%ebx,%%eax")
                   ENDCASE
    CASE f_mul:    cvf("MUL") // a := b * a; c := ?
                   writef("*n	imul %%ebx") // corrupts edx
                   ENDCASE
    CASE f_div:    cvf("DIV")  // a := b / a; c := ?
                   writef("*n	movl %%eax,%%ecx")
                   writef("*n	movl %%ebx,%%eax")
                   writef("*n	cltd")
                   writef("*n	idivl %%ecx")
                   ENDCASE
    CASE f_rem:    cvf("REM") // a := b REM a; c := ?
                   writef("*n	movl %%eax,%%ecx")
                   writef("*n	movl %%ebx,%%eax")
                   writef("*n	cltd")
                   writef("*n	idivl %%ecx")
                   writef("*n	movl %%edx,%%eax")
                   ENDCASE
    CASE f_add:    cvf("ADD") // a := b + a
                   writef("*n	addl %%ebx,%%eax")
                   ENDCASE
    CASE f_sub:    cvf("SUB") // a := b - a
                   writef("*n	subl %%ebx,%%eax")
                   writef("*n	negl %%eax")
                   ENDCASE
    CASE f_eq:     cvf("EQ") // a := b = a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	seteb %%al")
                   writef("*n	movzbl %%al,%%eax")
                   writef("*n	negl %%eax")
                   ENDCASE
    CASE f_ne:     cvf("NE") // a := b ~= a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	setneb %%al")
                   writef("*n	movzbl %%al,%%eax")
                   writef("*n	negl %%eax")
                   ENDCASE
    CASE f_ls:     cvf("LS") // a := b < a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	setlb %%al")
                   writef("*n	movzbl %%al,%%eax")
                   writef("*n	negl %%eax")
                   ENDCASE
    CASE f_gr:     cvf("GR") // a := b > a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	setgb %%al")
                   writef("*n	movzbl %%al,%%eax")
                   writef("*n	negl %%eax")
                   ENDCASE
    CASE f_le:     cvf("LE") // a := b <= a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	setleb %%al")
                   writef("*n	movzbl %%al,%%eax")
                   writef("*n	negl %%eax")
                   ENDCASE
    CASE f_ge:     cvf("GE") // a := b >= a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	setgeb %%al")
                   writef("*n	movzbl %%al,%%eax")
                   writef("*n	negl %%eax")
                   ENDCASE
    CASE f_eq0:    cvf("EQ0") // a := a = 0
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	seteb %%al")
                   writef("*n	movzbl %%al,%%eax")
                   writef("*n	negl %%eax")
                   ENDCASE
    CASE f_ne0:    cvf("NE0") // a := a ~= 0
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	setneb %%al")
                   writef("*n	movzbl %%al,%%eax")
                   writef("*n	negl %%eax")
                   ENDCASE
    CASE f_ls0:    cvf("LS0") // a := a < 0
                   writef("*n	sarl $31,%%eax")
                   ENDCASE
    CASE f_gr0:    cvf("GR0") // a := a > 0
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	setgb %%al")
                   writef("*n	movzbl %%al,%%eax")
                   writef("*n	negl %%eax")
                   ENDCASE
    CASE f_le0:    cvf("LE0") // a := a <= 0
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	setleb %%al")
                   writef("*n	movzbl %%al,%%eax")
                   writef("*n	negl %%eax")
                   ENDCASE
    CASE f_ge0:    cvf("GE0") // a := a >= 0
                   writef("*n	sarl $31,%%eax")
                   writef("*n	notl %%eax")
                   ENDCASE
    CASE f_lsh:    cvf("LSH") // a := b << a; b := ?	// ebx == a; eax == b
                   writef("*n	xchgl %%eax,%%ebx")
                   writef("*n	movl %%ebx,%%ecx")
                   writef("*n	cmpl $32,%%ecx")
                   writef("*n	sbbl %%ebx,%%ebx")		// set ebx to -1 or 0 (?)
                   writef("*n	andl %%ebx,%%eax")		// set eax to b or 0
                   writef("*n	sall %%cl,%%eax")		// now shift it
                   ENDCASE
    CASE f_rsh:    cvf("RSH") // a := b >> a; b := ?
                   writef("*n	xchgl %%eax,%%ebx")
                   writef("*n	movl %%ebx,%%ecx")
                   writef("*n	cmpl $32,%%ecx")
                   writef("*n	sbbl %%ebx,%%ebx")		// set ebx to -1 or 0 (?)
                   writef("*n	andl %%ebx,%%eax")		// set eax to b or 0
                   writef("*n	shrl %%cl,%%eax")		// now shift it
                   ENDCASE
    CASE f_and:    cvf("AND") // a := b & a 
                   writef("*n	andl %%ebx,%%eax") 
                   ENDCASE
    CASE f_or:     cvf("OR") // a := b | a 
                   writef("*n	orl %%ebx,%%eax") 
                   ENDCASE
    CASE f_xor:    cvf("XOR") // a := b NEQV a
                   writef("*n	xorl %%ebx,%%eax") 
                   ENDCASE
    CASE f_eqv:    cvf("EQV") // a := b EQV a 
                   writef("*n	xorl %%ebx,%%eax") 
                   writef("*n	notl %%eax") 
                   ENDCASE
    CASE f_gbyt:   cvf("GBYT") // a := b % a
                   writef("*n	movzbl (%%eax,%%ebx,4),%%eax") 
                   ENDCASE
    CASE f_xgbyt:  cvf("XGBYT") // a := a % b 
                   writef("*n	movzbl (%%ebx,%%eax,4),%%eax") 
                   ENDCASE
    CASE f_pbyt:   cvf("PBYT") // b % a := c
                   writef("*n	movb %%cl,(%%eax,%%ebx,4)") 
                   ENDCASE
    CASE f_xpbyt:  cvf("XPBYT") // a % b := c 
                   writef("*n	movb %%cl,(%%ebx,%%eax,4)") 
                   ENDCASE

// swb       Kn Ld K1 L1 ... Kn Ln   Binary chop switch, Ld default
    CASE f_swb:    cvswb()
                   ENDCASE

// swl       Kn Ld L1 ... Ln         Label vector switch, Ld default
    CASE f_swl:    cvswl()
                   ENDCASE
    CASE f_xch:    cvf("XCH") // swap a and b
                   writef("*n	xchgl %%eax,%%ebx")
                   ENDCASE
    CASE f_atb:    cvf("ATB") // b := a
                   writef("*n	movl %%eax,%%ebx")
                   ENDCASE
    CASE f_atc:    cvf("ATC") // c := a
                   writef("*n	movl %%eax,%%ecx")
                   ENDCASE
    CASE f_bta:    cvf("BTA") // a := b
                   writef("*n	movl %%ebx,%%eax")
                   ENDCASE
    CASE f_btc:    cvf("BTC") // c := b
                   writef("*n	movl %%ebx,%%ecx")
                   ENDCASE
    CASE f_atblp:  cvfp("ATBLP") // b := a; a := P!n
                   writef("*n	movl %%eax,%%ebx")
                   writef("*n	movl %n(%%ebp),%%eax", 4*pval)
                   ENDCASE
    CASE f_atblg:  cvfg("ATBLG") // b := a; a := G!n
                   writef("*n	movl %%eax,%%ebx")
                   writef("*n	movl %n(%%edi),%%eax", 4*gval)
                   ENDCASE
    CASE f_atbl:   cvfk("ATBL") // b := a; a := k
                   writef("*n	movl %%eax,%%ebx")
                   writef("*n	movl $%n,%%eax", kval)
                   ENDCASE
    CASE f_j:      cvfl("J") // jump to Ln
                   writef("*n	jmp L%c%n", modletter, lval)
                   ENDCASE
    CASE f_rtn:    cvf("RTN") // procedure return
              /*   writef("*n	movl %%ebp,%%ecx")
               *   writef("*n	mov 4(%%ecx),%%ebp")
               *   writef("*n	jmp **(%%ecx)")
               */
                   writef("*n	movl 4(%%ebp),%%ecx")
                   writef("*n	movl (%%ebp),%%ebp")
                   writef("*n	jmp **%%ecx")

                   ENDCASE
    CASE f_goto:   cvf("GOTO") // jump to a
                   writef("*n	jmp **%%eax")
                   ENDCASE
    CASE f_ikp:    cvfkp("IKP") // a := P!n + k; P!n := a
                   writef("*n	movl %n(%%ebp),%%eax", 4*pval)
                   TEST kval=1
                   THEN writef("*n	incl %%eax")
                   ELSE TEST kval=-1
                        THEN writef("*n	decl %%eax")
                        ELSE writef("*n	addl $%n,%%eax", kval)
                   writef("*n	movl %%eax,%n(%%ebp)", 4*pval)
                   ENDCASE
    CASE f_ikg:    cvfkg("IKG") // a := G!n + k; G!n := a
                   writef("*n	movl %n(%%edi),%%eax", 4*gval)
                   TEST kval=1
                   THEN writef("*n	incl %%eax")
                   ELSE TEST kval=-1
                        THEN writef("*n	decl %%eax")
                        ELSE writef("*n	addl $%n,%%eax", kval)
                   writef("*n	movl %%eax,%n(%%edi)", 4*gval)
                   ENDCASE
    CASE f_ikl:    cvfkl("IKL") // a := !Ln + k; !Ln := a
                   writef("*n	movl L%c%n,%%eax", modletter, lval)
                   TEST kval=1
                   THEN writef("*n	incl %%eax")
                   ELSE TEST kval=-1
                        THEN writef("*n	decl %%eax")
                        ELSE writef("*n	addl $%n,%%eax", kval)
                   writef("*n	movl %%eax,L%c%n", modletter, lval)
                   ENDCASE
    CASE f_ip:     cvfp("IP") // a := P!n + a; P!n := a
                   writef("*n	addl %n(%%ebp),%%eax", 4*pval)
                   writef("*n	movl %%eax,%n(%%ebp)", 4*pval)
                   ENDCASE
    CASE f_ig:     cvfg("IG") // a := G!n + a; G!n := a
                   writef("*n	addl %n(%%edi),%%eax", 4*gval)
                   writef("*n	movl %%eax,%n(%%edi)", 4*gval)
                   ENDCASE
    CASE f_il:     cvfl("IL") // a := !Ln + a; !Ln := a
                   writef("*n	addl L%c%n,%%eax", modletter, lval)
                   writef("*n	movl %%eax,L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jeq:    cvfl("JEQ") // Jump to Ln if b = a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	je L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jne:    cvfl("JNE") // Jump to Ln if b ~= a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	jne L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jls:    cvfl("JLS") // Jump to Ln if b < a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	jl L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jgr:    cvfl("JGR") // Jump to Ln if b > a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	jg L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jle:    cvfl("JLE") // Jump to Ln if b <= a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	jle L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jge:    cvfl("JGE") // Jump to Ln if b >= a
                   writef("*n	cmpl %%eax,%%ebx")
                   writef("*n	jge L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jeq0:   cvfl("JEQ0") // Jump to Ln if a = 0
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	je L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jne0:   cvfl("JNE0") // Jump to Ln if a ~= 0
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	jne L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jls0:   cvfl("JLS0") // Jump to Ln if a < 0
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	jl L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jgr0:   cvfl("JGR0") // Jump to Ln if a > 0
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	jg L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jle0:   cvfl("JLE0") // Jump to Ln if a <= 0
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	jle L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jge0:   cvfl("JGE0") // Jump to Ln if a >= 0
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	jge L%c%n", modletter, lval)
                   ENDCASE
    CASE f_jge0m:  cvfm("JGE0M") // Jump to Mn if a >= 0
                   writef("*n	orl %%eax,%%eax")
                   writef("*n	jge M%c%n", modletter, mval)
                   ENDCASE

    // The following five opcodes are never generated by
    // the BCPL compiler:

    CASE f_brk:    cvf("BRK") // Breakpoint instruction
                   writef("*n	unimplemented")
                   ENDCASE
    CASE f_nop:    cvf("NOP") // No operation
                   ENDCASE
    CASE f_chgco:  cvf("CHGCO") // Change coroutine
                   writef("*n	unimplemented")
                   ENDCASE
    CASE f_mdiv:   cvf("MDIV") // a := Muldiv(P!3, P!4, P!5) 
                   writef("*n	unimplemented")
                   ENDCASE
    CASE f_sys:    cvf("SYS") // System function
                   writef("*n	unimplemented")
                   ENDCASE


    CASE f_section:  cvfs("SECTION") // Name of section
                     FOR i = 0 TO charv%0 DO sectname%i := charv%i
                     ENDCASE
    CASE f_modstart: cvf("MODSTART") // Start of module  
                     sectname%0 := 0
                     ENDCASE
    CASE f_modend:   cvf("MODEND") // End of module 
                     modletter := modletter+1
                     ENDCASE
    CASE f_global:   cvglobal() // Global initialisation data
                     ENDCASE
    CASE f_string:   cvstring() // String constant
                     ENDCASE
    CASE f_const:    cvconst() // Large integer constant
                     ENDCASE
    CASE f_static:   cvstatic() // Static variable or table
                     ENDCASE
    CASE f_mlab:     cvfm("MLAB") // Destination of jge0m
                     writef("*nM%c%n:", modletter, mval)
                     ENDCASE
    CASE f_lab:      cvfl("LAB") // Program label
                     writef("*nL%c%n:", modletter, lval)
                     ENDCASE
    CASE f_lstr:     cvfm("LSTR") // a := Mn   (pointer to string)
                //   writef("*n	movl $M%c%n,%%eax", modletter, mval)
                     writef("*n	leal M%c%n,%%eax", modletter, mval)
                     writef("*n	shrl $2,%%eax")
                     ENDCASE
    CASE f_entry:    cventry() // Start of a function
                     ENDCASE


	//	Floating point ops added 8-September-2014:

    CASE f_float:    cvf("FLOAT")
                     loadb()
                     loada()
                     writef("*n	pushl %%eax")
                     writef("*n	filds (%%esp)") // st[0] := FLOAT a
                     writef("*n	fstps (%%esp)")
                     writef("*n	popl %%eax")    // a := FLOAT a
                     ENDCASE
    CASE f_fix:      cvf("FIX") // a := #ABS a
                     loadb()
                     loada()
                     writef("*n	pushl %%eax")
                     writef("*n	flds (%%esp)")  // st[0] := a
                     writef("*n	fistpl (%%esp)")
                     writef("*n	popl %%eax")    // a := FIX a
                     ENDCASE
    CASE f_fabs:     cvf("FABS") // a := #ABS a
                     loadb()
                     loada()
                     writef("*n	pushl %%eax")
                     writef("*n	flds (%%esp)")  // st[0] := a
                     writef("*n	fabs")          // st[0] := #ABS a
                     writef("*n	fstps (%%esp)")
                     writef("*n	popl %%eax")    // a := #ABS a
                     ENDCASE
    CASE f_fmul:     cvf("FMUL") // a := b #* a
                     loadb()
                     loada()
                     writef("*n	pushl %%ebx")
                     writef("*n	flds (%%esp)")  // st[0] := b
                     writef("*n	movl %%eax,(%%esp)")
                     writef("*n	fmuls (%%esp)") // st[0] := b #* a
                     writef("*n	fstps (%%esp)")
                     writef("*n	popl %%eax")    // a := b #+ a
                     ENDCASE
    CASE f_fdiv:     cvf("FDIV") // a := b #/ a
                     loadb()
                     loada()
                     writef("*n	pushl %%ebx")
                     writef("*n	flds (%%esp)")  // st[0] := b
                     writef("*n	movl %%eax,(%%esp)")
                     writef("*n	fdivs (%%esp)") // st[0] := b #/ a
                     writef("*n	fstps (%%esp)")
                     writef("*n	popl %%eax")    // a := b #/ a
                     ENDCASE
    CASE f_fadd:     cvf("FADD") // a := b #+ a
                     loadb()
                     loada()
                     writef("*n	pushl %%ebx")
                     writef("*n	flds (%%esp)")  // st[0] := b
                     writef("*n	movl %%eax,(%%esp)")
                     writef("*n	fadds (%%esp)") // st[0] := b #+ a
                     writef("*n	fstps (%%esp)")
                     writef("*n	popl %%eax")    // a := b #+ a
                     ENDCASE
    CASE f_fsub:     cvf("FSUB") // a := b #- a
                     loadb()
                     loada()
                     writef("*n	pushl %%ebx")
                     writef("*n	flds (%%esp)")  // st[0] := b
                     writef("*n	movl %%eax,(%%esp)")
                     writef("*n	fsubs (%%esp)") // st[0] := b #- a
                     writef("*n	fstps (%%esp)")
                     writef("*n	popl %%eax")    // a := b #- a
                     ENDCASE
    CASE f_fneg:     cvf("FNEG") // a := #- a
                     loadb()
                     loada()
                     writef("*n	pushl %%eax")
                     writef("*n	flds (%%esp)")  // st[0] := a
                     writef("*n	fchs")          // st[0] := #- a
                     writef("*n	fstps (%%esp)")
                     writef("*n	popl %%eax")    // a := #- a
                     ENDCASE
    CASE f_feq:      cvf("FEQ")
                     loadb()
                     loada()
                     writef("*n	pushl %%ebx")
                     writef("*n	flds (%%esp)")  // st[0] := b
                     writef("*n	pushl %%eax")
                     writef("*n	flds (%%esp)")  // st[0], st[1] = a, b
                     writef("*n	addl $8,%%esp")
                     writef("*n	fucomip %%st(1),%%st")
                     writef("*n	fstp %%st")
                     writef("*n	seteb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
    CASE f_fne:      cvf("FNE")
                     loadb()
                     loada()
                     writef("*n	pushl %%ebx")
                     writef("*n	flds (%%esp)")  // st[0] := b
                     writef("*n	pushl %%eax")
                     writef("*n	flds (%%esp)")  // st[0], st[1] = a, b
                     writef("*n	addl $8,%%esp")
                     writef("*n	fucomip %%st(1),%%st")
                     writef("*n	fstp %%st")
                     writef("*n	setneb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
    CASE f_fls:      cvf("FLS")
                     loadb()
                     loada()
                     writef("*n	pushl %%ebx")
                     writef("*n	flds (%%esp)")  // st[0] := b
                     writef("*n	pushl %%eax")
                     writef("*n	flds (%%esp)")  // st[0], st[1] = a, b
                     writef("*n	addl $8,%%esp")
                     writef("*n	fucomip %%st(1),%%st")
                     writef("*n	fstp %%st")
                     writef("*n	seta %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
    CASE f_fgr:      cvf("FGR")
                     loadb()
                     loada()
                     writef("*n	pushl %%eax")
                     writef("*n	flds (%%esp)")  // st[0] := a
                     writef("*n	pushl %%ebx")
                     writef("*n	flds (%%esp)")  // st[0], st[1] = b, a
                     writef("*n	addl $8,%%esp")
                     writef("*n	fucomip %%st(1),%%st")
                     writef("*n	fstp %%st")
                     writef("*n	seta %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
    CASE f_fle:      cvf("FLE")
                     loadb()
                     loada()
                     writef("*n	pushl %%ebx")
                     writef("*n	flds (%%esp)")  // st[0] := b
                     writef("*n	pushl %%eax")
                     writef("*n	flds (%%esp)")  // st[0], st[1] = a, b
                     writef("*n	addl $8,%%esp")
                     writef("*n	fucomip %%st(1),%%st")
                     writef("*n	fstp %%st")
                     writef("*n	setae %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
    CASE f_fge:      cvf("FGE")
                     loadb()
                     loada()
                     writef("*n	pushl %%eax")
                     writef("*n	flds (%%esp)")  // st[0] := a
                     writef("*n	pushl %%ebx")
                     writef("*n	flds (%%esp)")  // st[0], st[1] = b, a
                     writef("*n	addl $8,%%esp")
                     writef("*n	fucomip %%st(1),%%st")
                     writef("*n	fstp %%st")
                     writef("*n	setae %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
  }

//newline()
} REPEAT

AND writeq(str, a, b, c, d, e, f, g, h, i, j, k) BE RETURN

AND cvf(s)   BE writeq("# %s", s)
AND cvfp(s)  BE writeq("# %s P%n", s, rdp())
AND cvfkp(s) BE writeq("# %s K%n P%n", s, rdk(), rdp())
AND cvfg(s)  BE writeq("# %s G%n", s, rdg())
AND cvfkg(s) BE writeq("# %s K%n G%n", s, rdk(), rdg())
AND cvfkl(s) BE writeq("# %s K%n L%n", s, rdk(), rdl())
AND cvfpg(s) BE writeq("# %s P%n G%n", s, rdp(), rdg())
AND cvfk(s)  BE writeq("# %s K%n", s, rdk())
AND cvfw(s)  BE writeq("# %s W%n", s, rdw())
AND cvfl(s)  BE writeq("# %s L%n", s, rdl())
AND cvfm(s)  BE writeq("# %s M%n", s, rdm())

AND cvswl() BE
{ LET n = rdk()
  LET l = rdl()
  LET lab = nextlab()
  writeq("# SWL K%n L%n", n, l)
  writef("*n	orl %%eax,%%eax")
  writef("*n	jl L%c%n", modletter, l)
  writef("*n	cmpl $%n,%%eax", n)
  writef("*n	jge L%c%n", modletter, l)
  writef("*n	jmp **L%n(,%%eax,4)", lab)
  writef("*n	.data")
  writef("*n	.align 4")
  writef("*nL%n:", lab)
  FOR i = 1 TO n DO
  { writef("*n# L%n", rdl())
    writef("*n	.long L%c%n", modletter, lval)
  }
  writef("*n	.text")
}

AND cvswb() BE
{ LET n = rdk()
  LET l = rdl()
  writeq("# SWB K%n L%n", n, l)
  FOR i = 1 TO n DO 
  { LET k = rdk()
    LET l = rdl()
    writef("*n# K%n L%n", k, l)
    writef("*n	cmpl $%n,%%eax", k)
    writef("*n	je L%c%n", modletter, l)
  }
  writef("*n	jmp L%c%n", modletter, l)
}

AND cvglobal() BE
{ LET n = rdk()
  LET hwm = 0

  writef("*n# GLOBAL definitions: K%n*n", n)
  sectname := "G"
  IF sectname%0=0 FOR i = 0 TO 4 DO sectname%i := "prog"%i
  writef("*n.data*n")
//writef("*n.global %s*n", sectname)
  writef(".align 4*n")
//writef("%s:*n", sectname)

  FOR i = 0 TO n-1 DO
  { LET g = rdg()
	LET l = rdl()
	writef(".global G%n*n", g)
	writef(".equ G%n, L%c%n*n", g, modletter, l)
  }
  hwm := rdg()

}

AND rdchars() = VALOF
{ LET n = rdk()
  charv%0 := n
  FOR i = 1 TO n DO charv%i := rdc()
  RESULTIS n
}

AND cvstring() BE
{ LET lab = rdm()
  LET n = rdchars()
  // writeq("# STRING  M%n K%n", lab, n); FOR i = 1 TO n DO writef(" C%n", charv%i)
  writef("*n.data")
  writef("*n	.align 4")
  writef("*nM%c%n:", modletter, lab)
  FOR i = 0 TO n DO writef("*n	.byte %n", charv%i)
  writef("*n	.text")
}

AND cvconst() BE
{ LET lab = rdm()
  LET w = rdw()
//writef("# CONST   M%n W%n", lab, w)
  writef("*n.data")
  writef("*n	.align 4")
  writef("*nM%c%n:", modletter, lab)
  writef("*n	.long %n", w)
  writef("*n	.text")
}

AND cvstatic() BE
{ LET lab = rdl()
  LET n = rdk()
  writef("# STATIC  L%n K%n", lab, n)
  writef("*n.data")
  writef("*n	.align 4")
  writef("*nL%c%n:", modletter, lab)
  FOR i = 1 TO n DO { writef("*n# W%n", rdw())
                      writef("*n	.long %n", wval)
                    }
  writef("*n	.text")
}

AND cvfs(s) BE
{ LET n = rdchars()
//writef("*n# %s K%n", s, n)
//FOR i = 1 TO n DO writef(" C%n", charv%i)
  writef("*n# %s ", s)
  FOR i = 1 TO n DO wrch(charv%i)
}

AND cventry() BE
{ LET n = rdchars()
  LET op = rdf()
  LET lab = rdl()
  writef("*n# Entry to: %s()", charv)
  // writeq("# %s K%n", "ENTRY", n); FOR i = 1 TO n DO writef(" C%n", charv%i)
  // newline()
  TEST op=f_lab THEN writeq("# LAB     L%n*n", lab)
                ELSE writef("# Bad op F%n L%n*n", op, lab)
  writef("*nL%c%n:", modletter, lab)

/*	writef("*n	pop (%%ecx)")
 *	writef("*n	mov %%ebp,4(%%ecx)")	// P!0 := previous stackbase
 *	writef("*n	mov %%ecx,%%ebp")
 *	writef("*n	mov 12(%%ebp),%%eax")	// P!3 := arg2
 */

  writef("*n	movl 12(%%ebp),%%eax")	//	recover calling value of %eax...
}
